use std::cell::RefCell;
use std::rc::Rc;

pub trait CommandId {
    fn command_id(&self) -> u64;
}

impl CommandId for u64 {
    fn command_id(&self) -> u64 {
        *self
    }
}

pub trait WithCommandInfo {
    fn command_info(&self) -> &CommandInfo;
}

pub trait GetCommandInfo: CommandId {
    fn name(&self) -> &str;
}

pub struct CommandInfo {
    command_id: u64,
    name: String,
}

impl CommandInfo {
    pub fn new<TName: Into<String>>(command_id: u64, name: TName) -> Self {
        Self {
            command_id,
            name: name.into(),
        }
    }
}

impl CommandId for CommandInfo {
    fn command_id(&self) -> u64 {
        self.command_id
    }
}

impl GetCommandInfo for CommandInfo {
    fn name(&self) -> &str {
        &self.name
    }
}

impl<T: WithCommandInfo> CommandId for T {
    fn command_id(&self) -> u64 {
        self.command_info().command_id
    }
}

impl<T: WithCommandInfo> GetCommandInfo for T {
    fn name(&self) -> &str {
        &self.command_info().name
    }
}

#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub enum CommandStatus {
    Stopped,
    Running,
    Ended,
    Cancelling,
    Cancelled,
}

pub trait CommandParent {
    fn start_command(&mut self, cmd: CommandRef) -> bool;
}

pub trait Command: WithCommandInfo {
    fn reset(&mut self);
    fn status(&self) -> CommandStatus;
    fn execute(&mut self, &mut CommandParent) -> CommandStatus;
    fn cancel(&mut self);
}

pub type CommandRef<T = dyn Command> = Rc<RefCell<T>>;
