use super::{CommandContext, CommandImp, CommandRef, CommandResult, GetCommandInfo, CommandStatus, GenericCommand};

pub struct CommandGroupConfig {
    pub commands: Vec<CommandRef>,
}

impl CommandGroupConfig {
    pub fn new(commands: Vec<CommandRef>) -> Self {
        Self { commands }
    }
}

pub struct CommandGroupState {
    current_idx: usize,
    has_cancelled: bool,
}

impl Default for CommandGroupState {
    fn default() -> Self {
        CommandGroupState {
            current_idx: 0,
            has_cancelled: false,
        }
    }
}

pub type CommandGroup = GenericCommand<CommandGroupConfig>;

impl CommandImp for CommandGroupConfig {
    type State = CommandGroupState;

    fn init(&self, ctx: &mut CommandContext<Self>) {
        *ctx.state() = Default::default();
        debug!("CommandGroup {} init", ctx.name());
    }

    fn step(&self, ctx: &mut CommandContext<Self>) -> CommandResult {
        use self::CommandResult::*;
        use self::CommandStatus::*;
        let commands = &self.commands;
        if ctx.state().has_cancelled {
            debug!("CommandGroup step called after cancelled");
            return CommandResult::Cancel;
        }
        loop {
            if ctx.state().current_idx >= commands.len() {
                debug!("CommandGroup out of commands");
                return CommandResult::Done;
            }
            let mut command = commands[ctx.state().current_idx].borrow_mut();
            let status = command.execute(ctx.parent());
            match status {
                Ended => {
                    command.reset();
                    ctx.state().current_idx += 1;
                }
                Cancelled => {
                    command.reset();
                    ctx.state().has_cancelled = true;
                    return Cancel;
                }
                Running => return Continue,
                other => {
                    warn!("CommandGroup unhandled status {:?}", other);
                    return Continue;
                }
            }
        }
    }

    fn finish(&self, ctx: &mut CommandContext<Self>) {
        let commands = &self.commands;
        if ctx.status() == CommandStatus::Cancelling {
            if let Some(command) = commands.get(ctx.state().current_idx) {
                let mut command = command.borrow_mut();
                command.cancel();
                command.execute(ctx.parent());
                command.reset();
            }
        }
    }
}
