use super::{Command, CommandId, CommandParent, CommandRef, CommandStatus};

struct SchedulerCommandParent {
    start_queue: Vec<CommandRef>,
}

impl SchedulerCommandParent {
    fn new() -> Self {
        Self {
            start_queue: Vec::new(),
        }
    }
}

impl CommandParent for SchedulerCommandParent {
    fn start_command(&mut self, command: CommandRef) -> bool {
        self.start_queue.push(command);
        true
    }
}

pub struct Scheduler {
    running_commands: Vec<CommandRef>,
}

impl Scheduler {
    pub fn new() -> Self {
        Self {
            running_commands: Vec::new(),
        }
    }

    fn get_command<T: CommandId>(&self, id: T) -> Option<&CommandRef> {
        let id = id.command_id();
        for command in &self.running_commands {
            if command.borrow().command_info().command_id() == id {
                return Some(command);
            }
        }
        None
    }

    fn get_command_mut<T: CommandId>(&mut self, id: T) -> Option<&mut CommandRef> {
        let id = id.command_id();
        for command in &mut self.running_commands {
            if command.borrow().command_info().command_id() == id {
                return Some(command);
            }
        }
        None
    }

    pub fn start_command(&mut self, command: CommandRef) -> bool {
        let command_id = command.borrow().command_info().command_id();
        if self.is_running(command_id) {
            false
        } else {
            self.running_commands.push(command);
            true
        }
    }

    pub fn is_running<T: CommandId>(&self, id: T) -> bool {
        self.get_command(id).is_some()
    }

    pub fn cancel_command<T: CommandId>(&mut self, id: T) -> bool {
        self.get_command_mut(id).map_or(false, |command| {
            command.borrow_mut().cancel();
            true
        })
    }

    pub fn execute(&mut self) {
        use self::CommandStatus::*;
        let mut parent = SchedulerCommandParent::new();
        let exec_command =
            |command: &mut Command, parent: &mut CommandParent| match command.execute(parent) {
                Ended | Cancelled => {
                    command.reset();
                    true
                }
                Stopped | Running | Cancelling => false,
            };
        self.running_commands.drain_filter(|command| {
            let mut command = command.borrow_mut();
            exec_command(&mut *command, &mut parent)
        });
        while !parent.start_queue.is_empty() {
            let mut new_parent = SchedulerCommandParent::new();
            {
                let more_commands = parent.start_queue.into_iter().filter_map(|command_ref| {
                    let done = {
                        let mut command = command_ref.borrow_mut();
                        exec_command(&mut *command, &mut new_parent)
                    };
                    if !done {
                        Some(command_ref)
                    } else {
                        None
                    }
                });
                self.running_commands.extend(more_commands);
            }
            parent = new_parent;
        }
    }
}
