use std::cell::RefCell;
use std::rc::Rc;

use controller::PID_Controller;
use motor_controller::{MockMotorController, MotorControllerNewRef, MotorControllerRef};

pub struct Drive {
    left_motors: Vec<MotorControllerRef>,
    right_motors: Vec<MotorControllerRef>,
    left_controller: PID_Controller,
    right_controller: PID_Controller,
}

pub type DriveRef = Rc<RefCell<Drive>>;

fn create_motor_controllers<T: IntoIterator<Item = u16>>(ids: T) -> Vec<MotorControllerRef> {
    ids.into_iter().map(MockMotorController::new_ref).collect()
}

impl Drive {
    pub fn new_ref() -> Rc<RefCell<Self>> {
        let left_motors = create_motor_controllers(0u16..4);
        let right_motors = create_motor_controllers(4u16..8);
        let mut left_controller = PID_Controller::default();
        left_controller.input = Box::new(|| 0.5);
        let mut right_controller = PID_Controller::default();
        right_controller.input = Box::new(|| 0.5);
        right_controller.output = Box::new(left_motors[0].clone());
        let drive = Drive {
            left_motors,
            right_motors,
            left_controller,
            right_controller,
        };
        let drive = Rc::new(RefCell::new(drive));

        drive.borrow_mut().left_controller.output = {
            let drive = drive.clone();
            Box::new(move |out| drive.borrow().set_left_power(out))
        };
        drive.borrow_mut().right_controller.output = {
            let drive = drive.clone();
            Box::new(move |out| drive.borrow().set_right_power(out))
        };

        drive
    }

    pub fn set_left_power(&self, left_power: f64) {
        for motor in &self.left_motors {
            motor.set(left_power);
        }
    }
    pub fn set_right_power(&self, right_power: f64) {
        for motor in &self.right_motors {
            motor.set(right_power);
        }
    }

    pub fn drive_powers(&self, left_power: f64, right_power: f64) {
        self.set_left_power(left_power);
        self.set_right_power(right_power);
    }
}
